@@ -1,4 +1,8 @@
-"""Functions for post-processing ATS models."""
+"""Functions for post-processing ATS models.
+
+Authors: Pin Shuai (pin.shuai@usu.edu)
+
+"""
 import os, re, scipy, glob, itertools
 import numpy as np
 import pandas as pd
@@ -6,11 +10,14 @@ import h5py
 from math import sqrt
 from sklearn.metrics import mean_squared_error
 from scipy.stats.stats import pearsonr
+
 import matplotlib.pyplot as plt
+import matplotlib.dates as mdates
 import datetime as dt
 from datetime import datetime, timedelta
 import time
 from calendar import isleap
+
 import logging
 logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(name)s - %(levelname)s: %(message)s')
 import sys
@@ -26,6 +33,40 @@ rho_m = 55500 # moles/m^3, water molar density
 rho = 997
 g = 9.80665
 
+def get_8dayET(obs_et, simu_et, dates):
+    """convert daily ET to cumulative 8-day ET using MODIS data
+    Parameters:
+        obs_et, dataframe with single column
+        simu_et, dataframe with single column
+        
+    Returns:
+        df_8d_et, dataframe with two columns: simu ET and obs ET (both in mm/8d)
+    """
+    df_et = simu_et.loc[dates[0]:dates[1]].copy()
+
+    # time_ind = df_et.dropna().index
+    time_ind = obs_et.loc[dates[0]:dates[1], :].index
+
+    # df_8d_et = df_et.dropna().copy()
+    df_8d_et = df_et.reindex(time_ind).copy()
+    for i in np.arange(len(time_ind))[:]:
+        itime = time_ind[i]
+        if i == 0:
+            btime = df_et.index[0]
+        else:
+            btime = time_ind[i-1]
+
+        istart = list(df_et.index).index(btime)
+        iend = list(df_et.index).index(itime)    
+        # `loc` includes both ends, but `iloc` does not include end bound!
+        isum = df_et.iloc[istart:iend+1, :].sum(min_count = 1) # return NA is sum() contains NA
+        df_8d_et.loc[itime, :] = isum.values    
+        
+    df_8d_et['obs ET [mm/8d]'] = obs_et.loc[time_ind[0] : time_ind[-1], :].values
+    df_8d_et.rename(columns = {df_8d_et.columns[0]:'simu ET [mm/8d]'}, inplace=True)
+    
+    return df_8d_et
+
 def get_snow_cover(model_dir, **kwargs):
     """Calculate snow cover data from model output. Return a dataframe with snow cover percentage.
     Parameters: 
@@ -219,8 +260,8 @@ def combine_obs_output(old_dir, new_dir):
     df_old = load_waterBalance(old_dir, cumsum = False)
     df_new = load_waterBalance(new_dir, cumsum = False)
     
-    df_old_sub = df_old.loc[:, 'time [d]':'(Pr+SM-ET-Q) [m/d]']
-    df_new_sub = df_new.loc[:, 'time [d]':'(Pr+SM-ET-Q) [m/d]']
+    df_old_sub = df_old.loc[:, 'time [d]':'(Pr+SM-ET-Q) [m d^-1]']
+    df_new_sub = df_new.loc[:, 'time [d]':'(Pr+SM-ET-Q) [m d^-1]']
     
     assert(df_old_sub.shape[-1] == df_new_sub.shape[-1])
     df = pd.concat([df_old_sub, df_new_sub]) 
@@ -270,11 +311,13 @@ def plot_timestep(work_dir, fname_run_log=None):
 
     for line in searchlines[:]:
         if "Cycle" in line and "Time" in line: 
-            splitline = re.split("\s+|\x1b|,", line)
+            splitline = re.split(r"\s+|\x1b|,", line)
             # print(splitline)
+            # search for keywords "Cycle" and "Time", and "dt" in each line
             cycle_idx = splitline.index("Cycle")
             time_idx = splitline.index("Time")
             dt_idx = splitline.index("dt")
+            # find the position of the timestep, time and dt. This may change in different ATS version.
             istep = splitline[cycle_idx + 2]        
             itime = splitline[time_idx + 3]        
             idt = splitline[dt_idx + 3]
@@ -282,26 +325,37 @@ def plot_timestep(work_dir, fname_run_log=None):
             step_dt.append(istep)
             time_dt.append(itime)  
             
+        # find computing time and #of cores used 
+        # find #of cores used
         if "TimeMonitor" in line and "|" in line:
-            splitline = re.split("\s+|,", line)
+            splitline = re.split(r"\s+|,", line)
             ncores = splitline[-3]
-            
+        elif "TimeMonitor" in line:
+            splitline = re.split(r"\s+|,", line)
+            # print(splitline)
+            ncores = splitline[-3]
+        # find wallclock time. This may change in different ATS version.
         if "Simulation Driver  |  cycle" in line:
-            splitline = re.split("\s+|,", line) 
+            splitline = re.split(r"\s+|,", line) 
             walltime = float(splitline[4])/3600 #convert s to h
+        elif "4: solve" in line:
+            splitline = re.split(r"\s+|,", line) 
+            # print(splitline)
+            walltime = float(splitline[2])/3600
 
     dt = np.array([float(i) for i in dt])
     assert len(dt) > 0, "0 timestep found!"
     time_dt = np.array([float(i) for i in time_dt])
     nyear = np.ptp(time_dt)/365 # cnvt d to year
+    logging.info(f"Start day: {time_dt[0]}d; End day: {time_dt[-1]}d; Finished simulation time: {nyear*365}d")
     
     step_dt = [float(i) for i in step_dt] 
     df = pd.DataFrame(np.stack([step_dt, dt, time_dt]).T, columns = ['cycle', 'timestep [d]', 'times [d]'])
     ave_dt = np.array(dt).mean()*24 # convert from day to hour
-    
+    logging.info(f"Average timestep: {ave_dt} h")
+
     # plot
     fig,ax = plt.subplots(1,1, figsize=(8,6))
-
     ax.plot(step_dt, dt, 'k', lw = 0.5)
     ax.set_ylabel('Timestep (day)')
     ax.set_xlabel('Cycle')
@@ -309,8 +363,12 @@ def plot_timestep(work_dir, fname_run_log=None):
     
     try:
         walltime_per_year = walltime / nyear
-        ax.set_title(f"Cores: {ncores}; wallclock time [h]: {walltime:.2f}; ave ts [h]: {ave_dt:.2f}; ave walltime/yr [h]: {walltime_per_year:.2f}")
+        title=f"Cores: {ncores}; wallclock time [h]: {walltime:.2f}; ave ts [h]: {ave_dt:.2f}; ave walltime/yr [h]: {walltime_per_year:.2f}"
+        ax.set_title(title)
+        logging.info(title)
+        
     except:
+        logging.info("Could not find total wallclock time. Job may not finish.")
         pass
 
     ax1 = ax.twinx()
@@ -331,8 +389,12 @@ def rmLeapDays(time_ats, origin_date = "1980-01-01", noleap = True, freq = "D"):
             list of days or hours
         origin_date: datetime str in "%Y-%m-%d", default is 1980-1-1
             original datetime. In ATS, this depends on the starting date of Daymet. 
+        noleap: bool, default is True
+            if True, remove Dec31 from leap years
+        freq: str, default is "D"
+            frequency of time_ats. Either "D" or "H"
     Returns:
-        datetime array with leap days removed.
+        datetime array.
     """
     # round to the nearest int, this maybe necessary for
     # floating point erros
@@ -419,14 +481,14 @@ def get_subbasin_value(vis_data, varname, subbasin_cells, volume, times, weighte
     
     data = vis_data.getArray(varname)
     if 'transpiration' in varname or 'precip' in varname or 'snow-melt' in varname:
-        #convert m/s to mm/d
+        #convert m/s to mm d^-1
         data = data*86400*1000 
-        unit = '[mm/d]'
+        unit = '[mm d^-1]'
     elif 'flux' in varname:
         #convert mol/s to m^3/d
         data = data / rho_m *86400 
-        data = data/ volume *1000 # m3/d to mm/d
-        unit = '[mm/d]'
+        data = data/ volume *1000 # m3/d to mm d^-1
+        unit = '[mm d^-1]'
     elif "snow-water_equivalent" in varname:
         # convert m to mm
         data = data*1000 
@@ -471,21 +533,31 @@ def get_subbasin_value(vis_data, varname, subbasin_cells, volume, times, weighte
 
 
 def surfaceArea(model_dir):
-    """get surface area of ATS model"""
+    """get surface area [m^2] of ATS model. This assumes that *surface_data.h5 file is in the model directory.
+    Parameters:
+        model_dir: str
+            path to model directory
+    Returns:
+        surface area [m^2]
+    """
     h5_files = glob.glob(os.path.join(model_dir, "*surface_data.h5"))
     if len(h5_files) == 1:
         with h5py.File(h5_files[0]) as f:
-            a_key = list(f['surface-cell_volume.cell.0'].keys())[0] # pick any timestamp
-            surface_area = f['surface-cell_volume.cell.0'][a_key][:].sum() # m^2  
+            keys = list(f.keys())
+            # find group name starts with 'surface-cell_volume'
+            groupname = np.array(keys)[[key.startswith('surface-cell_volume') for key in keys]][0]
+            a_key = list(f[groupname].keys())[0] # pick any timestamp
+            surface_area = f[groupname][a_key][:].sum() # m^2  
     else:
-        raise KeyError("*surface_data.h5 could not be found!")
+        path = os.path.join(model_dir, "*surface_data.h5")
+        raise KeyError(f"*surface_data.h5 could not be found in {path}!")
     return surface_area
     
 def load_waterBalance(model_dir, WB_filename = "water_balance.dat", timestep =
                       'D', UTC_time = None, resample_freq = None, canopy =
-                      True, origin_date = "1980-01-01", noleap = True, cumsum = True,
+                      True, origin_date = "1980-01-01", noleap = True, 
                       restart_dir = None, out_file = None, plot = False,
-                      subcatchment=False, catchment_area=None,
+                      catchment_area=None, sep = ',', domain_names=None,
                       **kwargs):
     """read ATS output files, new dataframe format
     Parameters:
@@ -505,30 +577,35 @@ def load_waterBalance(model_dir, WB_filename = "water_balance.dat", timestep =
             Original datetime in Daymet forcing time column
         noleap: bool, default is True
             Assume model ignores the leap year to be consistent with Daymet product
-        cumsum: bool,
-            If true, generate cumsum for vairables
         restart_dir: list or None
             If list, merge with restarted model output
         out_file: str or None
             If string, save water balance in a csv file.
         plot: bool,
             If true, plot water balance plots
+        catchment_area: float or None
+            If float, use this value as catchment area. Otherwise, calculate from model output (requires "*surface_data.h5" file!).
+        sep: str
+            delimiter used in the water balance file. Default is ','
+        domain_names: list or None
+            If list, use these names as domain names. Otherwise, use all domains.
     Returns:
-        datetime, data, and dataframe(datetime, data)
+        dataframe of model variables and/or a resampled dataframe.
     
     """
-    if subcatchment:
+    # get surface area
+    if catchment_area is not None:
         surface_area = catchment_area
     else:
         surface_area = surfaceArea(model_dir)
     assert(surface_area is not None)
 
-    df = load_output(model_dir, WB_filename, timestep, origin_date, **kwargs)
-    
+    df = load_output(model_dir, WB_filename, timestep, origin_date, noleap=noleap, sep=sep, **kwargs)
+
     if restart_dir is not None:
         assert(type(restart_dir) is list)
         for idir in restart_dir:
-            df_rst = load_output(idir, WB_filename, timestep, origin_date,
+            df_rst = load_output(idir, WB_filename, timestep, origin_date, noleap=noleap, sep=sep, 
                                  **kwargs)
             if df.shape[-1] != df_rst.shape[-1]:
                 logging.warning("waterbalance.out file columns does not match! Trying merging anyway.")
@@ -546,11 +623,15 @@ def load_waterBalance(model_dir, WB_filename = "water_balance.dat", timestep =
 
     if "river discharge [mol d^-1]" in df.columns:
         df["river discharge [m^3/d]"] = df['river discharge [mol d^-1]'] / rho_m
-        df["river discharge [m/d]"] = df['river discharge [m^3/d]'] / surface_area
+        df["river discharge [m d^-1]"] = df['river discharge [m^3/d]'] / surface_area
 
     if "net groundwater flux [mol d^-1]" in df.columns:
         df["net groundwater flux [m^3/d]"] = df['net groundwater flux [mol d^-1]'] / rho_m
-        df["net groundwater flux [m/d]"] = df['net groundwater flux [m^3/d]'] / surface_area
+        df["net groundwater flux [m d^-1]"] = df['net groundwater flux [m^3/d]'] / surface_area
+
+    if "net exchange flux [mol d^-1]" in df.columns:
+        df["net exchange flux [m^3/d]"] = df['net exchange flux [mol d^-1]'] / rho_m
+        df["net exchange flux [m d^-1]"] = df['net exchange flux [m^3/d]'] / surface_area
 
     # convert units
     if timestep == 'H':
@@ -572,34 +653,20 @@ def load_waterBalance(model_dir, WB_filename = "water_balance.dat", timestep =
     df['snow precipitation [mm d^-1]'] = df['snow precipitation [m d^-1]']*1000
     df['snowmelt [mm d^-1]'] = df['snowmelt [m d^-1]']*1000
     df['watershed boundary discharge [m^3/d]'] = df['watershed boundary discharge [mol d^-1]'] / rho_m
-    df['watershed boundary discharge [m/d]'] = df['watershed boundary discharge [m^3/d]'] / surface_area
+    df['watershed boundary discharge [m d^-1]'] = df['watershed boundary discharge [m^3/d]'] / surface_area
     df["total evapotranspiration [mm d^-1]"] = df["total evapotranspiration [m d^-1]"] *1000         
     df['snow water content [m]'] = df['snow water content [mol]'] / rho_m / surface_area
     df['surface water content [m]'] = df['surface water content [mol]'] / rho_m / surface_area
-    df['subsurface water content [m]'] = df['subsurface water content [mol]'] / rho_m / surface_area        
+    df['subsurface water content [m]'] = df['subsurface water content [mol]'] / rho_m / surface_area     
+    # df['exfiltration [m d^-1]'] = df['exfiltration [mol d^-1]'] / rho_m / surface_area
     if not canopy:            
         df['total water content [m]'] = df['surface water content [m]'] + df['subsurface water content [m]'] + df['snow water content [m]']
     else:
         df["canopy water content [m]"] = df["canopy water content [mol]"]/ rho_m / surface_area
-        df['(Ps_thru-SM-Es) [m/d]'] = df['snow to surface [m SWE d^-1]'] - df['snowmelt [m d^-1]'] -df['snow evaporation [m d^-1]']
         df['total water content [m]'] = df['surface water content [m]'] + df['subsurface water content [m]'] + df['snow water content [m]'] + df["canopy water content [m]"]
-      
-    if "net groundwater flux [mol d^-1]" in df.columns:
-        df['(Pr+SM-ET-Q-gw) [m/d]'] = df['rain precipitation [m d^-1]'] + df['snowmelt [m d^-1]'] -\
-                            df['total evapotranspiration [m d^-1]'] - df['watershed boundary discharge [m/d]'] -\
-                            df['net groundwater flux [m/d]']
-        df['(Pr+Ps-ET-Q-gw) [m/d]'] = df['rain precipitation [m d^-1]'] + df['snow precipitation [m d^-1]'] -\
-                            df['total evapotranspiration [m d^-1]'] - df['watershed boundary discharge [m/d]'] -\
-                            df['net groundwater flux [m/d]']
-    # calculate water balance
-    df['(Pr+SM-ET-Q) [m/d]'] = df['rain precipitation [m d^-1]'] + df['snowmelt [m d^-1]'] -\
-                        df['total evapotranspiration [m d^-1]'] - df['watershed boundary discharge [m/d]']
-    df['(Pr+Ps-ET-Q) [m/d]'] = df['rain precipitation [m d^-1]'] + df['snow precipitation [m d^-1]'] -\
-                        df['total evapotranspiration [m d^-1]'] - df['watershed boundary discharge [m/d]']
-    df['(Ps-SM) [m/d]'] = df['snow precipitation [m d^-1]'] - df['snowmelt [m d^-1]']
-
+    
     if resample_freq is not None:
-        variables = ['rain precipitation [m d^-1]','snow precipitation [m d^-1]','snowmelt [m d^-1]', 'watershed boundary discharge [m/d]', 'watershed boundary discharge [m^3/d]', 'total evapotranspiration [m d^-1]', 'max ponded depth [m]', 'SWE [m]']
+        variables = ['rain precipitation [m d^-1]','snow precipitation [m d^-1]','snowmelt [m d^-1]', 'watershed boundary discharge [m d^-1]', 'watershed boundary discharge [m^3/d]', 'total evapotranspiration [m d^-1]', 'max ponded depth [m]', 'SWE [m]']
         df_rs = df.loc[:, variables].copy()
         logging.info(f"Resample to {resample_freq} frequency")
         df_rs = df_rs.resample(resample_freq).mean()
@@ -607,43 +674,111 @@ def load_waterBalance(model_dir, WB_filename = "water_balance.dat", timestep =
         
         if UTC_time is not None:
             df_rs.index = df_rs.index.shift(UTC_time, freq = 'H')     
-    
-    if cumsum:
-        
-        if "net groundwater flux [mol d^-1]" in df.columns:
-            df["cum_groundwater flux [m]"] = df['net groundwater flux [m/d]'].cumsum()
-            df['cum_(Pr+Ps-ET-Q-gw) [m]'] = df['(Pr+Ps-ET-Q-gw) [m/d]'].cumsum()
-        if canopy:
-            df['cum canopy mass change [m]'] = df["canopy water content [m]"].cumsum()
-            df['cum_snow evaporation [m]'] = df['snow evaporation [m d^-1]'].cumsum()
-            df["cum_snow to surface [m]"] = df['snow to surface [m SWE d^-1]'].cumsum()
-            df['cum_(Ps_thru-SM-Es) [m]'] = df['(Ps_thru-SM-Es) [m/d]'].cumsum()
-            df['cum_(P-Pt+S-St-Ec-drainage) [m]'] = (df['rain precipitation [m d^-1]'] - df["water to surface [m d^-1]"] + df["snow precipitation [m d^-1]"] - df["snow to surface [m SWE d^-1]"] - df["canopy evaporation [m d^-1]"] - df["canopy drainage [m d^-1]"]).cumsum()
-        df["cum_rain precipitation [m]"] = df['rain precipitation [m d^-1]'].cumsum()
-        df["cum_snow precipitation [m]"] = df['snow precipitation [m d^-1]'].cumsum()
-        df["cum_snowmelt [m]"] = df['snowmelt [m d^-1]'].cumsum()
-        df["cum_ET [m]"] = df["total evapotranspiration [m d^-1]"].cumsum()
-        df['cum_overland flux [m]'] = df['watershed boundary discharge [m/d]'].cumsum()
-        df['cum_(Pr+SM-ET-Q) [m]'] = df['(Pr+SM-ET-Q) [m/d]'].cumsum()
-        df['cum_(Pr+Ps-ET-Q) [m]'] = df['(Pr+Ps-ET-Q) [m/d]'].cumsum()
-        df['cum_(Ps-SM) [m]'] = df["(Ps-SM) [m/d]"].cumsum()
-        df['cum snow mass change [m]'] = df['snow water content [m]'] - df['snow water content [m]'].values[0]
-        df['cum water mass change [m]'] = df['total water content [m]'] - df['total water content [m]'].values[0]
-
-        if "net groundwater flux [mol d^-1]" in df.columns:
-            df['water mass error [m]'] = df['cum_(Pr+Ps-ET-Q-gw) [m]'] - df['cum water mass change [m]']
-        else:
-            df['water mass error [m]'] = df['cum_(Pr+Ps-ET-Q) [m]'] - df['cum water mass change [m]']
-
-        if not canopy:
-            df['snow mass error [m]'] = df['cum_(Ps-SM) [m]'] - df['cum snow mass change [m]']
-        else:
-            df['snow mass error [m]'] = df['cum_(Ps_thru-SM-Es) [m]'] - df['cum snow mass change [m]']
-            
+                
     if plot:
-        
-        fig, axes = plt.subplots(3, 1, figsize=(8,8))
+        # plot water balance
+        if domain_names is None:
+            if canopy:
+                domain_names = ['global', 'canopy', 'snow', 'surface', 'subsurface']
+            else:
+                domain_names = ['global', 'snow', 'surface', 'subsurface']
 
+        def cumu_plot(domain, df, ax):
+            """Plot cumulative fluxes to check water balance at each domain. 
+            The steps follow Ethan's notebook here: https://github.com/amanzi/ats-demos/blob/master/05_ecohydrology/water_balance_pt_canopy.ipynb
+
+            Parameters:
+            -----------
+                domain: str
+                    'global' or 'canopy' or 'snow' or 'surface' or 'subsurface'
+                df: pd.DataFrame
+                    dataframe containing fluxes
+                ax: matplotlib.axes
+                    axes to plot on
+
+            Water balances:
+                - global: dWC/dt = Pr + Ps - ET - Q or dWC/dt = Pr + Ps - ET - Q - Gw (Pr: rain precipitation, Ps: snow precipitation, ET: evapotranspiration, Q: overland flow, Gw: net_groundwater_flux)
+                - canopy: dWC/dt = I - D - Ec (I: canopy interception, D: drainage, Ec: canopy evaporation)
+                - snow: dWC/dt = S_thru+D - SM - Es (S_thru: snow throughfall, D: drainage, SM: snowmelt, Es: sublimation or snow evaporation)
+                - surface: dWC/dt = R_thru+D - SM - E - Q - Inf (R_thru: rain throughfall, D: drainage, SM: snowmelt, E: surface evaporation, Q: overland flow, Inf: net_infiltration or net_exchange)
+                - subsurface: dWC/dt = Inf - T or I - T - Gw (Inf: net_infiltration or net_exchange, T: transpiration, Gw: net_groundwater_flux))  
+
+            Variables: 
+                - Gw or net_groundwater_flux: the amount of water entering or leaving the subsurface domain though the external_sides boundary. Typically, this would be zero if the subsurface domain is closed. i.e., no flow boundary conditions at bottom and sides. However, if subwatershed is included and there is flow exchange between subwatersheds, then this value will be non-zero.
+                - Inf or net_infiltration/net_exchange: the net exchange amount of water entering or leaving the surface domain. Positive [+] values indicate water entering the surface domain or leaving the subsurface domain. Negative [-] values indicate water leaving the surface domain or entering the subsurface domain.
+            """
+            time = df.index
+            fluxes = {}
+            if domain == "global":
+                fluxes['Pr'] = df['rain precipitation [m d^-1]']
+                fluxes['Ps'] = df['snow precipitation [m d^-1]']
+                fluxes['ET'] = df['total evapotranspiration [m d^-1]']
+                fluxes['Q'] = df['watershed boundary discharge [m d^-1]']
+                if "net groundwater flux [mol d^-1]" in df.columns:
+                    fluxes['Gw'] = df['net groundwater flux [m d^-1]']
+                    net_fluxes = fluxes['Pr'] + fluxes['Ps'] - fluxes['ET'] - fluxes['Q'] - fluxes['Gw']
+                else:
+                    net_fluxes = fluxes['Pr'] + fluxes['Ps'] - fluxes['ET'] - fluxes['Q']
+                water_content_change = df['total water content [m]'] - df['total water content [m]'].values[0]
+
+            if domain == 'canopy':
+                fluxes['I'] = df['canopy interception [m d^-1]']
+                fluxes['D'] = df['canopy drainage [m d^-1]']
+                fluxes['Ec'] = df['canopy evaporation [m d^-1]']
+                net_fluxes = fluxes['I'] - fluxes['D'] - fluxes['Ec']
+                water_content_change = df['canopy water content [m]'] - df['canopy water content [m]'].values[0]
+
+            if domain == 'snow':
+                fluxes['S_thru+D'] = df['snow to surface [m SWE d^-1]']
+                fluxes['Es'] = df['snow evaporation [m d^-1]']
+                fluxes['SM'] = df['snowmelt [m d^-1]']
+                net_fluxes = fluxes['S_thru+D'] - fluxes['Es'] - fluxes['SM']
+                water_content_change = df['snow water content [m]'] - df['snow water content [m]'].values[0]
+
+            if domain == 'surface':
+                fluxes['R_thru+D'] = df['water to surface [m d^-1]']
+                fluxes['SM'] = df['snowmelt [m d^-1]']
+                fluxes['E'] = df['surface evaporation [m d^-1]']
+                fluxes['Q'] = df['watershed boundary discharge [m d^-1]']
+                fluxes['Inf'] = -df['net exchange flux [m d^-1]']
+                net_fluxes = fluxes['R_thru+D'] + fluxes['SM'] - fluxes['E'] - fluxes['Q'] - fluxes['Inf']
+                water_content_change = df['surface water content [m]'] - df['surface water content [m]'].values[0]
+
+            if domain == 'subsurface':
+                fluxes['Inf'] = -df['net exchange flux [m d^-1]']
+                fluxes['T'] = df['transpiration [m d^-1]']
+                if "net groundwater flux [mol d^-1]" in df.columns:
+                    fluxes['Gw'] = df['net groundwater flux [m d^-1]']
+                    net_fluxes = fluxes['Inf'] - fluxes['T'] - fluxes['Gw']
+                else:
+                    net_fluxes = fluxes['Inf'] - fluxes['T']
+                water_content_change = df['subsurface water content [m]'] - df['subsurface water content [m]'].values[0]
+
+            error = np.cumsum(net_fluxes) - water_content_change
+            max_error = max(abs(error.max()), abs(error.min()))
+
+            for k,v in fluxes.items():
+                ax.plot(time, np.cumsum(v), label=k)
+
+            ax.plot(time, water_content_change, 'crimson', label='dW/dt')
+            ax.plot(time, np.cumsum(net_fluxes), 'k--', label='net fluxes')
+            ax.plot(time, error, '-.', color='grey', label='error')
+            ax.set_ylabel('CumFlux [m]')
+            ax.set_title(f'{domain} water balance (max error = {max_error*1000:.2f} mm)')
+            
+            ax.set_xlim(df.index[0], df.index[-1])
+            ax.set_xlabel('')
+            ax.legend(loc='upper left', fontsize = 10, bbox_to_anchor = (1.0,1.0), frameon = False)
+            ax.minorticks_off()
+            ax.xaxis.set_major_locator(mdates.YearLocator())
+            ax.xaxis.set_major_formatter(mdates.DateFormatter('%Y'))
+            # if domain != "subsurface":
+            #     ax.xaxis.set_ticklabels([])
+
+
+        fig, axes = plt.subplots(len(domain_names) + 1, 1, figsize=(8, max(6, 3 * len(domain_names))), sharex=True, dpi=150)
+ 
+        # plot incoming and outgoing fluxes
         ax = axes[0]
         df.plot(y=["rain precipitation [m d^-1]", "snowmelt [m d^-1]" ], color
                 = colors[:2],
@@ -651,14 +786,14 @@ def load_waterBalance(model_dir, WB_filename = "water_balance.dat", timestep =
         ax1 = ax.twinx()
         if "river discharge [mol d^-1]" in df.columns:
             df.plot(y=["total evapotranspiration [m d^-1]", 
-                   "river discharge [m/d]"], color =
+                   "river discharge [m d^-1]"], color =
                 colors[2:4], ax = ax1)
         else:
             df.plot(y=["total evapotranspiration [m d^-1]", 
-                   "watershed boundary discharge [m/d]"], color =
+                   "watershed boundary discharge [m d^-1]"], color =
                 colors[2:4], ax = ax1)
-        ax.set_ylabel('Incoming flux [m/d]')
-        ax1.set_ylabel('Outgoing flux [m/d]')
+        ax.set_ylabel('Incoming flux [m d^-1]')
+        ax1.set_ylabel('Outgoing flux [m d^-1]')
         ax.legend(loc='upper left', fontsize = 10, bbox_to_anchor = (0.0,1.3), frameon = False)
         ax1.legend(loc='upper left', fontsize = 10, bbox_to_anchor = (0.5,1.3),
                   frameon = False)
@@ -667,36 +802,11 @@ def load_waterBalance(model_dir, WB_filename = "water_balance.dat", timestep =
         ax.set_xlim(df.index[0], df.index[-1])
         ax.set_xlabel('')
 
-        ax = axes[1]
-        if "net groundwater flux [mol d^-1]" in df.columns:
-            df.plot(y=["cum_rain precipitation [m]", "cum_snow precipitation [m]", "cum_ET [m]", 
-                   "cum_overland flux [m]", "cum_groundwater flux [m]", 'cum_(Pr+Ps-ET-Q-gw) [m]'], ax = ax)
-        else:
-            df.plot(y=["cum_rain precipitation [m]", "cum_snow precipitation [m]", "cum_ET [m]", 
-                   "cum_overland flux [m]", 'cum_(Pr+Ps-ET-Q) [m]'], ax = ax)
-        df.plot(y = ['cum water mass change [m]', 'water mass error [m]'], style = ['-.','-'], color = ['k', 'grey'], ax = ax)
-        ax.set_ylabel('CumFlux [m]')
-        ax.legend(loc='upper left', fontsize = 10, bbox_to_anchor = (1.0,1.15), frameon = False)
-        max_error = max(abs(df['water mass error [m]'].max()), abs(df['water mass error [m]'].min()))
-#         print(max_error)
-        ax.set_title(f'max error = {max_error*1000:.2f} mm')
-        ax.xaxis.set_ticklabels([])
-        ax.set_xlim(df.index[0], df.index[-1])
-        ax.set_xlabel('')
-
-        ax = axes[2]
-        if not canopy:
-            df.plot(y=["cum_snow precipitation [m]", "cum_snowmelt [m]", 'cum_(Ps-SM) [m]'], ax = ax)
-            df.plot(y=['cum snow mass change [m]', 'snow mass error [m]'], style = ['-.','-'], color = ['k', 'grey'], ax = ax)
-        else:
-            df.plot(y=[ "cum_snow to surface [m]", "cum_snowmelt [m]", 'cum_snow evaporation [m]', 'cum_(Ps_thru-SM-Es) [m]'], ax=ax)
-            df.plot(y=['cum snow mass change [m]', 'snow mass error [m]'], style = ['-.','-'], color = ['k', 'grey'], ax = ax)
-        ax.set_ylabel('CumFlux [m]')
-        ax.legend(loc='upper left', fontsize = 10, bbox_to_anchor = (1.0,1.1), frameon = False)
-        max_error = max(abs(df['snow mass error [m]'].max()), abs(df['snow mass error [m]'].min()))
-        ax.set_title(f'max error = {max_error*1000:.2f} mm')
-        ax.set_xlim(df.index[0], df.index[-1])  
-        ax.set_xlabel('')
+        # plot all other domains
+        for i, domain in enumerate(domain_names):
+            ax = axes[i + 1]
+            cumu_plot(domain, df, ax)
+        fig.tight_layout()
     
     if UTC_time is not None:
         df.index = df.index.shift(UTC_time, freq = 'H')         
@@ -710,7 +820,7 @@ def load_waterBalance(model_dir, WB_filename = "water_balance.dat", timestep =
         return df
 
 def load_output(model_dir, WB_filename, timestep = 'D', origin_date =
-                '1980-01-01', **kwargs):
+                '1980-01-01', noleap=True, sep = ',', **kwargs):
     """read ATS output files
     Parameters:
         model_dir: str
@@ -718,24 +828,34 @@ def load_output(model_dir, WB_filename, timestep = 'D', origin_date =
         WB_filename: str
             Waterbalance output filename, e.g. "water_balance.dat"
         timestep: str
-            timestep string, 'D', 'H'
+            timestep string, 'D', 'H'. Default is 'D'
         origin_date: str,
             origin of model time
+        noleap: bool
+            remove leap days or not. Default is True.
+        sep: str
+            delimiter used in the observation file. Default is ','
+        kwargs: dict
+            additional arguments passed to rmLeapDays()
     Returns:
         datetime, data, and dataframe(datetime, data)
     
     """
-    df = pd.read_csv(os.path.join(model_dir, WB_filename), comment='#')
+    df = pd.read_csv(os.path.join(model_dir, WB_filename), sep=sep,  comment='#')
+    # Assert that the DataFrame is not empty
+    assert not df.empty, "The DataFrame is empty. Double check the path to the water balance file."
+    assert 'time [d]' in df.columns or 'time [h]' in df.columns, "The DataFrame does not have either 'time [d]' or 'time [s]' column. Double check the df and water balance file. Did you use the correct separator in pd.read_csv()? Defaults to comma."
+
     if timestep == 'D':
         # get datetime, convert seconds to days
-        datetime_noLeap = rmLeapDays(df['time [d]'], freq='D', origin_date =
-                                     origin_date, **kwargs)
+        datetime_ats = rmLeapDays(df['time [d]'], freq='D', origin_date =
+                                     origin_date, noleap=noleap, **kwargs)
     elif timestep == 'H':
-        datetime_noLeap = rmLeapDays(df['time [h]'], freq='H', origin_date =
-                                     origin_date, **kwargs)
+        datetime_ats = rmLeapDays(df['time [h]'], freq='H', origin_date =
+                                     origin_date, noleap=noleap, **kwargs)
     else:
         raise ValueError(f"Freq: {timestep} is not supported!")
-    df['datetime'] = datetime_noLeap
+    df['datetime'] = datetime_ats
     df.set_index('datetime', inplace=True, drop=True)
     
     return df
