@@ -29,12 +29,79 @@ def gantt_plot(ranges, **kwargs):
     """
     fig,ax = plt.subplots(1,1, figsize=(8,6))
 
+
+    ax = ax.xaxis_date()
+    ax = plt.hlines(ranges.index,
+                    mdates.date2num(ranges['start']),
+                    mdates.date2num(ranges['end']), **kwargs)
+    fig.tight_layout()
+
+
+def gantt_plot_from_df(dfs, ax=None, **kwargs):
+    """
+    Plot Gantt plot given start and end date.
+
+    Parameters:
+        dfs : DataFrame or list of DataFrames
+            Dataframe(s) with index as name/id and two separate columns ['start', 'end'].
+            If a single DataFrame is provided, it should have a MultiIndex with the first level representing the name/id. Or, the DataFrame should have a single column representing the name/id.
+            If a list of DataFrames is provided, each DataFrame should have a single column representing the name/id.
+
+        ax : matplotlib.axes.Axes, optional
+            The axes on which to plot the Gantt chart. If not provided, a new figure and axes will be created.
+
+        **kwargs : dict, optional
+            Additional keyword arguments to be passed to the `plt.hlines` function.
+
+    Returns:
+        fig : matplotlib.figure.Figure or None
+            The Figure object if `ax` is None, otherwise None.
+
+        ax : matplotlib.axes.Axes or None
+            The Axes object if `ax` is None, otherwise None.
+    """
+    ndf = len(dfs)
+    if ndf == 1 and isinstance(dfs.index, pd.MultiIndex):
+        ranges = []
+        for index in dfs.index.levels[0]:
+            df = dfs.loc[index]
+            df.index = pd.to_datetime(df.index)
+            start_dates, end_dates = utils.mark_start_end(df)
+            for i in range(len(start_dates)):
+                ranges.append((index, start_dates[i].to_pydatetime(), end_dates[i].to_pydatetime()))
+    elif ndf == 1 and isinstance(dfs[0], pd.DataFrame):
+        ranges = []
+        df = dfs[0]
+        # df = df[df.columns[0]]
+        start_dates, end_dates = utils.mark_start_end(df)
+        # print(len(start_dates))
+        for i in range(len(start_dates)):
+            ranges.append((df.columns[0], start_dates[i].to_pydatetime(), end_dates[i].to_pydatetime()))
+
+    if ndf > 1 and isinstance(dfs[0], pd.DataFrame):
+        ranges = []
+        for i in range(1, ndf):
+            df = dfs[i]
+            start_dates, end_dates = utils.mark_start_end(df)
+            for i in range(len(start_dates)):
+                ranges.append((df.columns[0], start_dates[i].to_pydatetime(), end_dates[i].to_pydatetime()))
+    # print(ranges)
+    ranges = pd.DataFrame(ranges, columns=['id', 'start', 'end']) 
+    ranges.set_index('id', inplace=True)   
+    # print(ranges.head())
+    if ax is None:
+        fig, ax = plt.subplots(1, 1, figsize=(8, 6))
+
     ax = ax.xaxis_date()
     ax = plt.hlines(ranges.index, 
-                    mdates.date2num(ranges['start']), 
-                    mdates.date2num(ranges['end']), **kwargs)
+              mdates.date2num(ranges['start']), 
+              mdates.date2num(ranges['end']), **kwargs)
+    plt.grid(linestyle='-.')
     fig.tight_layout()  
 
+    if ax is None:
+        return fig, ax
+
 def plot_FDC(dfs, labels, colors, linestyles=None, start_date=None, end_date=None,
              time_index=None, rank_method = 'average', var=None, ax=None, **kwargs):
     """plot flow duration curve for discharge time series.
@@ -63,6 +130,9 @@ def plot_FDC(dfs, labels, colors, linestyles=None, start_date=None, end_date=Non
             Column variable if provided
         ax, axis
             axis for plotting
+
+    Returns:
+        fig, ax if ax is None.
     """
     ndf = len(dfs)
 
@@ -192,7 +262,19 @@ def corrMatrix_plot(df, **kwargs):
                )    
     plt.title('Correlation Matrix', fontsize = 14)
 def varsComp_plot(dfs, variables, labels, colors, linestyles=None, axes=None, **kwargs):
-    """compare variables across different model runs"""
+    """compare variables across different model runs.
+    Parameters:
+    ---
+        dfs, list of dataframes
+        variables, list of variables
+        labels, list of labels for legend
+        colors, list of colors for line plot
+
+    Returns:
+    ---
+        None or (fig, axes)
+    
+    """
     nvar = len(variables)
     if axes is None:
         fig, axes = plt.subplots(nvar,1, figsize=(8,3*nvar), sharex = True)
@@ -205,6 +287,8 @@ def varsComp_plot(dfs, variables, labels, colors, linestyles=None, axes=None, **
     if nvar > 1:
         for ax, var in zip(axes, variables):
             for i in np.arange(len(dfs)):
+                idf = dfs[i][var]
+                assert idf.empty is False, f"{var} does not exist in {i+1}th df!"
                 ax.plot(var, linestyles[i], color = colors[i], data = dfs[i], label = labels[i], **kwargs)
                 ax.set_ylabel(var, fontsize = 12)
                 ax.set_xlabel('')
@@ -213,6 +297,8 @@ def varsComp_plot(dfs, variables, labels, colors, linestyles=None, axes=None, **
         ax = axes
         var = variables[0]
         for i in np.arange(len(dfs)):
+            idf = dfs[i][var]
+            assert idf.empty is False, f"{var} does not exist in {i+1}th df!"
             ax.plot(var, linestyles[i], color = colors[i], data = dfs[i], label = labels[i], **kwargs)
             ax.set_ylabel(var, fontsize = 12)
             ax.set_xlabel('')
@@ -360,10 +446,10 @@ def quantile_plot(data, quantiles= [0.25, 0.5, 0.75], axis= 1, arr_index =
 
 def one2one_plot(df_obs, df_simu, metrics=["R^2"], show_metrics=True,
                 ax=None, equal_aspect=True, show_density=False,
-                 decompose_KGE=False, start_date=None, end_date=None, **kwargs):
+                 decompose_KGE=False, start_date=None, end_date=None, dropzero=False, **kwargs):
     """One to One plot with a line.
     Parameters:
-        df_obs, df_simu are Pandas series.
+        df_obs, df_simu are Pandas time series with datetime as index.
         metric, list or 'all'
             available metrics are: pearsonr, R2, RMSE, KGE,...
         show_metrics, bool
@@ -381,6 +467,8 @@ def one2one_plot(df_obs, df_simu, metrics=["R^2"], show_metrics=True,
             The start of the datetime index. e.g., "2016-10-01"
         end_date, str
             The end of the datetime index. e.g., "2020-10-01"
+        dropzero, bool
+            Drop zero values from the data before calculating metrics. This is useful when artifacts are present in the data.
     Returns:
         A one to one plot with metrics.
     """
@@ -394,12 +482,12 @@ def one2one_plot(df_obs, df_simu, metrics=["R^2"], show_metrics=True,
         df_simu = df_simu.loc[start_date:end_date]
 
     metric_dict, df = utils.get_metrics(df_obs.index, df_obs.values, 
-                 df_simu.index, df_simu.values, metrics = metrics)
+                 df_simu.index, df_simu.values, metrics = metrics, dropzero=dropzero)
     if "mKGE" in metrics or 'KGE' in metrics: 
         if decompose_KGE is True:
             KGE_metric, _ = utils.get_metrics(df_obs.index, df_obs.values, 
-                 df_simu.index, df_simu.values, metrics = ['mKGE'], return_all
-                                         = True)
+                 df_simu.index, df_simu.values, metrics = ['mKGE'], 
+                 return_all= True, dropzero=dropzero)
             print(f"mKGE: {KGE_metric['mKGE'][0]}, cc:  {KGE_metric['mKGE'][1]}, alpha:  {KGE_metric['mKGE'][2]}, beta:  {KGE_metric['mKGE'][3]}")
     
     max_ = max(df['obs'].values.max(), df['simu'].values.max())
