@@ -9,6 +9,7 @@ logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(name)s - %(level
 from modvis.ATSutils import rmLeapDays
 # import fiona, shapely
 from datetime import datetime
+import geopandas as gpd
 
 rho_m = 55500 # moles/m^3, water molar density
 rho = 997
@@ -125,9 +126,7 @@ def plot_water_content(vis_data,
     except:
         return ax, tpc
 
-def plot_column_head(vis_data, 
-                     origin_date='1980-01-01', col_ind = 0, plot = True,
-                     ax=None):
+def plot_column_head(vis_data, origin_date='1980-01-01', col_ind = 0, cell_id = None, infer_col_from_cell = False, plot = True, ax=None):
     """plot variable in a single column over time in the subsurface.
     Parameters:
         vis_data, object from xdmf.VisFile()
@@ -135,6 +134,10 @@ def plot_column_head(vis_data,
             model start origin time. Defaults to 1980-1-1
         col_ind, int
             column index for plotting head
+        cell_id, int
+            cell id in a column. This is used for inferring the column index. The cell id can be obtained by viewing visdump file in Paraview/Visit. This can be useful if col_ind is not known. Defaults to None.
+        infer_col_from_cell, bool
+            infer column index from a given cell id if True. Defaults to False.            
         plot, bool
             plot the head if true
     Returns:
@@ -147,6 +150,11 @@ def plot_column_head(vis_data,
     times = get_time(vis_data, origin_date=origin_date)
     # times = vis_data.times
     # datetime = rmLeapDays(times)
+
+    # obtain column index from given cell id
+    if infer_col_from_cell and cell_id is not None:
+        col_ind = np.where(map == cell_id)[0][0]
+
     dat = vis_data.getArray("pressure")
 
     iz_coord = ordered_centroids[col_ind, :, -1]
@@ -289,7 +297,7 @@ def plot_gw_surface(visfile, origin_date="1980-01-01", time_slice = -1, return_h
         except:
             return ax, tpc
 
-def plot_column_data(vis_data, var_name, origin_date='1980-01-01', col_ind = 0, cmap = None, ylabel = None, plot_contour = False, contour_spacing = 0.01, levels = None, logx = False, ax=None):
+def plot_column_data(vis_data, var_name, origin_date='1980-01-01', col_ind = 0, cell_id = None, infer_col_from_cell = False, cmap = None, ylabel = None, plot_contour = False, contour_spacing = 0.01, levels = None, logx = False, ax=None):
     """plot variable in a single column over time in the subsurface.
 
     Parameters:
@@ -299,7 +307,11 @@ def plot_column_data(vis_data, var_name, origin_date='1980-01-01', col_ind = 0,
         origin_date, str
             model start origin time. Defaults to 1980-1-1
         col_ind, int
-            column index for plotting head
+            column index for plotting variables such as pressure. Defaults to 0.
+        cell_id, int
+            cell id in a column. This is used for inferring the column index. The cell id can be obtained by viewing visdump file in Paraview/Visit. This can be useful if col_ind is not known. Defaults to None.
+        infer_col_from_cell, bool
+            infer column index from a given cell id if True. Defaults to False.
         cmap, str
             colormap for plotting
         plot_contour, bool
@@ -318,6 +330,10 @@ def plot_column_data(vis_data, var_name, origin_date='1980-01-01', col_ind = 0,
     ordered_centroids = vis_data.centroids
     map = vis_data.map
 
+    # obtain column index from given cell id
+    if infer_col_from_cell and cell_id is not None:
+        col_ind = np.where(map == cell_id)[0][0]
+    
     # times = vis_data.times
     times = get_time(vis_data, origin_date=origin_date)
     dat = vis_data.getArray(var_name)
@@ -370,7 +386,11 @@ def plot_column_data(vis_data, var_name, origin_date='1980-01-01', col_ind = 0,
         return fig, ax
     except:
         return
-def plot_layer_data(vis_data, var_name, origin_date = '1980-01-01', layer_ind = 0, time_slice = -1, cmap = "turbo", colorbar = True, clabel = None, ax = None, log = False, vmin = None, vmax = None, linthresh = 0.01, linscale = 0.01, **kwargs):
+def plot_layer_data(vis_data, var_name, origin_date = '1980-01-01', layer_ind = 0, 
+                    time_slice = -1, cmap = "viridis", colorbar = True, clabel = None, 
+                    ax = None, log = False, vmin = None, vmax = None, robust = False, linthresh = 0.01, 
+                    linscale = 0.01, mixed_element = False, data_lim = None,
+                    **kwargs):
     """plot variable in a single layer in the subsurface.
     Parameters:
         vis_data, ats_xdmf.VisFile object
@@ -379,7 +399,8 @@ def plot_layer_data(vis_data, var_name, origin_date = '1980-01-01', layer_ind =
         origin_date, str
             original date used in the model
         layer_ind, int, 0-indexed
-            layer id with 0 being on top and -1 on bottom. Note the actual layers are ordered from bottom up.
+            layer id with 0 being on top and -1 on bottom. 
+            Note the actual layers in the mesh are ordered from bottom up with 0 being at the bottom.
         time_slice, int, 0-indexed
             time index-0,1,...,-1
         cmap, str
@@ -391,55 +412,94 @@ def plot_layer_data(vis_data, var_name, origin_date = '1980-01-01', layer_ind =
         ax, axis handel. Default is creating one.
         log, bool
             set plot to log scale if True
+        vmin, float
+            minimum value for color scale
+        vmax, float
+            maximum value for color scale
+        robust, bool
+            whether to use robust scaling (i.e., percentile [2,98] scaling)
         linthresh, float
             keyword for SymLogNorm        
         linscale, float
             see keyword for SymLogNorm
-        
+        mixed_element, bool
+            whether to plot mixed element meshes
+        data_lim, tuple
+            data limits for plotting
+
     Returns:
-        fig, ax   
+        fig, ax, gdf if mixed_element else fig, ax, tpc
+
     """
-    ordered_centroids = vis_data.centroids
-    vertex_xyz = vis_data.vertex_xyz
-    conn = vis_data.conn
+    if mixed_element:
+        mesh_polygons = vis_data.mesh_polygons
+        remapping = vis_data.remapping
+    else:
+        vertex_xyz = vis_data.vertex_xyz
+        conn = vis_data.conn
     map = vis_data.map
-#     times = vis_data.times
-#     datetime = rmLeapDays(times)
+
     times, time_idx = get_time(vis_data, time_slice, origin_date = origin_date)
     
-#     dat = vis_data.getArray(var_name)
     if isinstance(var_name, str):
-        dat = vis_data.getArray(var_name)
+        if mixed_element:
+            dat = vis_data.getArray(var_name)[:][:, remapping[True], :]
+        else:
+            dat = vis_data.getArray(var_name)
     elif isinstance(var_name, (np.ndarray, np.generic)):
+        # TODO: need to add sth for mixed_element
         dat = var_name
         assert(dat.shape[0] == len(times))
     else:
         raise RuntimeError("Must provide string of variable or np.array of data!")
     
-    # layer ordered from bottom to top
+    # layer ordered from bottom to top (i.e., 0--bottom, -1--top)
+    # here we reverse the order so 0--top, -1--bottom for easier selection
     layers = np.arange(map.shape[-1])[::-1]
     ilayer = layers[layer_ind]
-#     icoord = ordered_centroids[:, ilayer, :]
-    icells = map[:, ilayer].flatten()
-    # idat = dat[time_idx, icells]
+
     idat = dat[time_idx, :, ilayer]
-    iconn = conn[icells, -3:]
+
+    if data_lim is not None:
+        if data_lim[0] is not None:
+            idat[idat < data_lim[0]] = np.nan
+        if data_lim[1] is not None:
+            idat[idat > data_lim[1]] = np.nan
+
+    if not mixed_element:
+        icells = map[:, ilayer].flatten()
+        iconn = conn[icells, -3:]
+
     if vmin is None:
-        vmin = np.nanmin(dat)
-    if vmax is None: 
-        vmax = np.nanmax(dat)
+        vmin = np.nanmin(idat)
+    if vmax is None:
+        vmax = np.nanmax(idat)
+    if robust:
+        vmin, vmax = np.nanpercentile(idat, [2, 98])
+
     if ax == None:
         fig, ax = plt.subplots(1,1, figsize=(8, 4))    
     ax.set_aspect('equal')
-    if log:
-        tpc = ax.tripcolor(vertex_xyz[:,0], vertex_xyz[:,1], iconn, 
-                       facecolors= idat, linewidth=0.01, cmap = cmap, 
-                       norm=matplotlib.colors.SymLogNorm(linthresh = linthresh, linscale = linscale, vmin=vmin, vmax=vmax),
-                       **kwargs)
+
+    if mixed_element:
+        gdf = gpd.GeoDataFrame({'value': idat, 'geometry': mesh_polygons})
+        gdf.plot(column='value', ax=ax, cmap=cmap, 
+                vmin=vmin, vmax=vmax,
+                 legend=False)
+
+        # Create a normalizer and ScalarMappable
+        norm = matplotlib.colors.Normalize(vmin=vmin, vmax=vmax)
+        sm = matplotlib.cm.ScalarMappable(cmap=cmap, norm=norm)        
     else:
-        tpc = ax.tripcolor(vertex_xyz[:,0], vertex_xyz[:,1], iconn,
-                facecolors= idat, cmap = cmap, linewidth=0.01, 
-                vmin = vmin, vmax = vmax, **kwargs)
+        if log:
+            tpc = ax.tripcolor(vertex_xyz[:,0], vertex_xyz[:,1], iconn, 
+                        facecolors= idat, linewidth=0.01, cmap = cmap, 
+                        norm=matplotlib.colors.SymLogNorm(linthresh = linthresh, linscale = linscale, vmin=vmin, vmax=vmax),
+                        **kwargs)
+        else:
+            tpc = ax.tripcolor(vertex_xyz[:,0], vertex_xyz[:,1], iconn,
+                    facecolors= idat, cmap = cmap, linewidth=0.01, 
+                    vmin = vmin, vmax = vmax, **kwargs)
     
     ax.set_title(f"Time: {times[time_idx].date()}; Layer: {layer_ind+1}")
     ax.set_xlabel("Easting [m]")
@@ -449,23 +509,30 @@ def plot_layer_data(vis_data, var_name, origin_date = '1980-01-01', layer_ind =
             clabel = var_name 
         elif clabel is None:
             clabel = ''
-#         clabel = var_name
-        cb = plt.colorbar(tpc)
-        cb.ax.set_ylabel(clabel, labelpad=0.3)
-#     plt.tight_layout()
+        if mixed_element:
+            cb = plt.colorbar(sm, ax=ax, label=clabel)
+        else:
+            cb = plt.colorbar(tpc)
+            cb.ax.set_ylabel(clabel, labelpad=0.3)
     
-    if ax is None:
-        return fig, ax, tpc
+    if mixed_element:
+        try:
+            return fig, ax, gdf
+        except:
+            return ax, gdf
     else:
-        return ax, tpc
+        try:
+            return fig, ax, tpc
+        except:
+            return ax, tpc
 
 def plot_surface_data(vis_data, var_name,
                       origin_date="1980-01-01", time_slice = -1,
-                      facecolors = None,  subset= False, subset_idx = None,
+                      facecolors = None, cmap = 'viridis', subset= False, subset_idx = None,
                       colorbar = True, clabel = None, title = True, ax = None,
                       log = False, vmin = None, vmax = None, robust = False,
                       contour = False, nlevel = 5, linthresh = 0.01, linscale =
-                      0.01, data_lim = None, **kwargs):
+                      0.01, data_lim = None, mixed_element = False, **kwargs):
     """plot variable on the surface.
     Parameters:
         vis_data, ats_xdmf.VisFile object
@@ -501,18 +568,31 @@ def plot_surface_data(vis_data, var_name,
         data_lim, None or list
             Only show data within this range if provided. Out ranged values are
             marked as Nans. 
+        mixed_element, bool
+            whether the mesh is mixed element or not. Default to False
     Returns:
-        fig, ax   
+        fig, ax, gdf if mixed_element else fig, ax, tpc
+
     """
-    vertex_xyz = vis_data.vertex_xyz
-    conn = vis_data.conn
-    volume = vis_data.volume
+
+    if mixed_element:
+        mesh_polygons = vis_data.mesh_polygons
+        remapping = vis_data.remapping
+    else:
+        vertex_xyz = vis_data.vertex_xyz
+        conn = vis_data.conn
+        volume = vis_data.volume
+        iconn = conn[:, -3:]
 
     times, time_idx = get_time(vis_data, time_slice, origin_date=origin_date)
     
     if isinstance(var_name, str):
-        dat = vis_data.getArray(var_name)
+        if mixed_element:
+            dat = vis_data.getArray(var_name)[:][:, remapping[True], 0]
+        else:
+            dat = vis_data.getArray(var_name)
     elif isinstance(var_name, (np.ndarray, np.generic)):
+        # TODO: need to add sth for mixed_element
         dat = var_name
         assert(dat.shape[0] == len(times))
     else:
@@ -535,7 +615,7 @@ def plot_surface_data(vis_data, var_name,
         else:
             logging.info("No unit convertion.")
             unit = ''      
-    
+    # subset timestamp
     idat = dat[time_idx, :]
 
     if data_lim is not None:
@@ -543,16 +623,18 @@ def plot_surface_data(vis_data, var_name,
             idat[idat < data_lim[0]] = np.nan
         if data_lim[1] is not None:
             idat[idat > data_lim[1]] = np.nan
-        
 
-    iconn = conn[:, -3:]
-    
+    if vmin is None:
+        vmin = np.nanmin(idat)
+    if vmax is None:
+        vmax = np.nanmax(idat)
     if robust:
         vmin, vmax = np.nanpercentile(idat, [2, 98])
         
     if facecolors is not None:
         colors = facecolors
     else:
+        # TODO: add remapping for mixed element mesh
         colors = idat
         
     if subset:
@@ -563,30 +645,39 @@ def plot_surface_data(vis_data, var_name,
         fig, ax = plt.subplots(1,1, figsize=(8, 4))
     ax.set_aspect('equal')
     
-    if log and not contour: 
-        # use SymLogNorm to normalize in both positive and negative directions.
-        tpc = ax.tripcolor(vertex_xyz[:,0], vertex_xyz[:,1], iconn, 
-                       facecolors= colors, linewidth=0.01,norm=matplotlib.colors.SymLogNorm(linthresh=linthresh, linscale=linscale, vmin=vmin, vmax=vmax), **kwargs)
-    elif not contour: 
-        tpc = ax.tripcolor(vertex_xyz[:,0], vertex_xyz[:,1], iconn, 
-                       facecolors= colors, linewidth=0.01, vmin=vmin, vmax=vmax, **kwargs)
-    elif contour:
-        
-        if isinstance(nlevel, int):
-            levels = np.linspace(np.floor(idat.min()), np.ceil(idat.max()), nlevel)
-        elif isinstance(nlevel, list):
-            levels = nlevel
-        else:
-            raise RuntimeError("Must provide level info for contour! Can be a list of levels of number of levels")
-        # get color value at vertex
-        val = get_vertex_value_from_cell(vertex_xyz, iconn, idat)
+    if mixed_element:
+        gdf = gpd.GeoDataFrame({'value': idat, 'geometry': mesh_polygons})
+        gdf.plot(column='value', ax=ax, cmap=cmap, 
+                 vmin=vmin, vmax=vmax,
+                 legend=False)
+        # Create a normalizer and ScalarMappable
+        norm = matplotlib.colors.Normalize(vmin=vmin, vmax=vmax)
+        sm = matplotlib.cm.ScalarMappable(cmap=cmap, norm=norm)
 
-        tpc = ax.tricontourf(vertex_xyz[:,0], vertex_xyz[:,1], iconn, 
-                       val, levels = levels, extend = 'both', **kwargs)  
-        ax.tricontour(vertex_xyz[:,0], vertex_xyz[:,1], iconn, 
-                       val, colors = 'k', linewidths = 0.5, levels = levels, extend = 'both', **kwargs)  
+    else:
+        if log and not contour: 
+            # use SymLogNorm to normalize in both positive and negative directions.
+            tpc = ax.tripcolor(vertex_xyz[:,0], vertex_xyz[:,1], iconn, 
+                        facecolors= colors, linewidth=0.01,norm=matplotlib.colors.SymLogNorm(linthresh=linthresh, linscale=linscale, vmin=vmin, vmax=vmax), **kwargs)
+        elif not contour: 
+            tpc = ax.tripcolor(vertex_xyz[:,0], vertex_xyz[:,1], iconn, 
+                        facecolors= colors, linewidth=0.01, vmin=vmin, vmax=vmax, **kwargs)
+        elif contour:
+            
+            if isinstance(nlevel, int):
+                levels = np.linspace(np.floor(idat.min()), np.ceil(idat.max()), nlevel)
+            elif isinstance(nlevel, list):
+                levels = nlevel
+            else:
+                raise RuntimeError("Must provide level info for contour! Can be a list of levels of number of levels")
+            # get color value at vertex
+            val = get_vertex_value_from_cell(vertex_xyz, iconn, idat)
+
+            tpc = ax.tricontourf(vertex_xyz[:,0], vertex_xyz[:,1], iconn, 
+                        val, levels = levels, extend = 'both', **kwargs)  
+            ax.tricontour(vertex_xyz[:,0], vertex_xyz[:,1], iconn, 
+                        val, colors = 'k', linewidths = 0.5, levels = levels, extend = 'both', **kwargs)  
    
-    
     if title is False:
         titles = ''
     else:
@@ -597,15 +688,25 @@ def plot_surface_data(vis_data, var_name,
     if colorbar == True:
         if clabel is None and isinstance(var_name, str):
             clabel = var_name + ' '+ unit
-        cb = plt.colorbar(tpc, 
-#                           extend = "both"
-                         )
-        cb.ax.set_ylabel(clabel, labelpad=5)
+
+        if mixed_element:
+            cb = plt.colorbar(sm, ax=ax, label=clabel)
+        else:
+            cb = plt.colorbar(tpc, 
+    #                           extend = "both"
+                            )
+            cb.ax.set_ylabel(clabel, labelpad=5)
     
-    try:
-        return fig, ax, tpc
-    except:
-        return ax, tpc
+    if mixed_element:
+        try:
+            return fig, ax, gdf
+        except:
+            return ax, gdf
+    else:
+        try:
+            return fig, ax, tpc
+        except:
+            return ax, tpc
     
 # def plot_riverbed(source, vertex_xyz, triangles, rivers=None, dist_to_river = 200, plot = True):
 #     """find riverbed region given river shapefile or index."""
@@ -649,3 +750,4 @@ def plot_surface_data(vis_data, var_name,
 #         return river_idx, shps
 #     except:
 #         return river_idx
+
