@@ -7,7 +7,7 @@ import pandas as pd
 import h5py
 from math import cos, sin, asin, sqrt, radians
 from sklearn.metrics import mean_squared_error
-from scipy.stats.stats import pearsonr
+from scipy.stats import pearsonr, mannwhitneyu
 import matplotlib.pyplot as plt
 from datetime import datetime, timedelta
 import time
@@ -23,6 +23,70 @@ import sys
 # import objectivefunctions as ofs
 import modvis.objectivefunctions as ofs
 
+def insertNA(df, dt_col, freq=None):
+    '''
+    Automatically insert NAs for missing data.
+    Arguments:
+      - df. A dataframe contains the missing data.
+      - dt_col. Column name in the df which contains time series.
+    '''
+    df[dt_col] = pd.to_datetime(df[dt_col])
+    ## remove duplicates in the datetime index
+#     df = df[~df[dt_col].duplicated()]
+    df.drop_duplicates(subset=dt_col, inplace=True)
+    df.sort_values(by = dt_col, inplace = True)
+    # drop na
+    # df = df.loc[df.index.notnull()]
+    df.dropna(subset=[dt_col], inplace=True)
+    ## create a new index
+    start_time = df[dt_col].iloc[0]
+    end_time = df[dt_col].iloc[-1]
+    # ic(start_time)
+    # ic(end_time)
+    
+    # find frequency
+    if freq is None:
+        ## get the highest frequency in sec
+        freq_list = df[dt_col].diff().dt.seconds.value_counts()
+
+        # the first freq with the most samples on the list may not be the highest freq
+        if freq_list.index[0] == freq_list.index.min():
+            freq = freq_list.index[0]
+        else:
+            freq = freq_list.index.min()
+    logging.info(f'Found the highest frequency: {freq} sec')
+    new_index = pd.date_range(start_time, end_time, freq=pd.DateOffset(seconds = freq),
+                              name = dt_col)
+    ## reindex and reset index
+    df = df.set_index(dt_col).reindex(new_index).reset_index()
+
+    return df
+
+def days_between(d0,d1, noLeap=False):
+    """calculates days between two dates
+        Parameters:
+            d0, datetime.date()
+                starting date
+            d1, datetime.date() object
+                ending date
+            noLeap, bool
+                whether to consider leap years. 
+                If True, only 365 days are considered each year.
+                
+        Returns:
+            days in between dates.
+    """
+    delta = d1-d0
+    if noLeap:
+        # remove one day from leap year 
+        start_year = d0.year
+        end_year = d1.year
+        leap_years = list(range(start_year + (4 - start_year % 4), end_year + 1, 4))
+        
+        return delta.days - len(leap_years)
+    else:
+        return delta.days
+
 def load_nwis(service='dv', parameterCd='00060', SI_unit=True, insert_NA=True, rm_tz=True, **kwargs):
     """Load USGS streaflow from NWIS.
         Parameters:
@@ -44,9 +108,9 @@ def load_nwis(service='dv', parameterCd='00060', SI_unit=True, insert_NA=True, r
     """
     df = nwis.get_record(service=service, parameterCd=parameterCd, **kwargs)
     assert len(df) > 0, "Returned empty df!"
-    
+
     # remove outliers and nan index
-    df[df.iloc[:, 0] < 0 ] = np.nan
+    df[df.loc[:, f'{parameterCd}_Mean'] < 0 ] = np.nan
     df = df[df.index.notnull()]
     
     if SI_unit:
@@ -68,28 +132,33 @@ def load_nwis(service='dv', parameterCd='00060', SI_unit=True, insert_NA=True, r
         
     return df
 
+def mark_start_end(df, freq='1D'):
+    """Mark the start and end of the dateframe time series. This can be used in combination with
+    utils.gantt_plot().
+    Parameters:
+        df: pandas.DataFrame
+            dataframe with datetime index and at least one column of data
+        freq: str
+            frequency of the data. Defaults to '1D' (daily).
+    Returns:
+        start_dates: list of datetime
+        end_dates: list of datetime
+    """
 
-def mark_start_end(df, label = None):
-    """Mark the start and end of the dateframe time series."""
-    ranges = []
-    # Start of good data block defined by a number preceeded by a NaN
-    start_mark = (df.notnull() & df.shift().isnull())
-    start = df[start_mark].index
+    # drop NaN data first
+    df.dropna(inplace=True)
+    # create a mask for missing data (assumes data is spaced evenly)
+    mask = df.index.to_series().diff() > pd.Timedelta(freq)    
 
-    # End of good data block defined by a number followed by a Nan
-    end_mark = (df.notnull() & df.shift(-1).isnull())
-    end = df[end_mark].index  
-    
-    for s, e in zip(start, end):
-        if label is None:
-            ranges.append((s, e))
-            # ranges = pd.DataFrame(ranges, columns=['start', 'end']) 
-        else:
-            ranges.append((label, s, e))
-            # ranges = pd.DataFrame(ranges, columns=['id', 'start', 'end']) 
-            # ranges.set_index('id', inplace = True)
-    
-    return ranges
+    # find the date before the mask (i.e. new starting date), will get new ending date
+    end_idx = list(np.where(mask.values)[0] - 1)
+    # find the start and end dates for each available time period
+    start_dates = [df.index[0]] + df.index[mask].to_list()
+    end_dates = df.index[end_idx].to_list() + [df.index[-1]]
+
+    assert len(start_dates) == len(end_dates)
+
+    return start_dates, end_dates
 
 def K_perm_conversion(K=None, k=None):
     """convert between hydraulic conductivity (K) and permeability (k)
@@ -300,8 +369,9 @@ def print_xml(file):
     print(''.join(file))
 
 def get_metrics(obs_t, obs, simu_t, simu, metrics = 'all', start_date = None,
-                end_date = None, epsilon=0, **kwargs):
-    """compute R2 and RMSE between simulated and observed
+                end_date = None, epsilon=0, dropzero=False, **kwargs):
+    """
+    Calculate metrics between simulated and observed data.
     Parameters:
         obs_t: list or array
             observed datetime series
@@ -312,12 +382,17 @@ def get_metrics(obs_t, obs, simu_t, simu, metrics = 'all', start_date = None,
         simu: list or array
             simulated numeric sereis
         start_date: datetime object
-            starting date for RMSE and R^2 calculation
+            starting date for metric calculation
         end_date: datetime object
-            ending date for RMSE and R^2 calculation
+            ending date for metric calculation
         epsilon: float
-            small number added to the data to avoid zero issue when using log transformation (e.g., logNSE). Recommended value is 1/100 of the mean observed flow (see Santos et al., 2018 HESS for discussion).
-
+            small number added to the data to avoid zero issue when using log transformation (e.g., logNSE). 
+            Recommended value is 1/100 of the mean observed flow (see Santos et al., 2018 HESS for discussion).
+        dropzero: bool
+            if True, drop zero values in observed data. This is useful for dataframe with artifacts of zero values 
+            (e.g., when using the restarted output from ATS).
+        **kwargs: keyword arguments
+            keyword arguments {return_all = True} for KGE components calculation
     Returns:
         dict of metrics and dataframe used.
     
@@ -338,6 +413,10 @@ def get_metrics(obs_t, obs, simu_t, simu, metrics = 'all', start_date = None,
     df2.dropna(inplace = True)
     
     df = df1.merge(df2, how='inner', on = 'datetime')
+
+    if dropzero:
+        # drop rows with any zero values
+        df = df[(df != 0).all(1)]
     
     if start_date != None or end_date != None:
         df = df[df.datetime < end_date].copy()
@@ -364,6 +443,7 @@ def get_metrics(obs_t, obs, simu_t, simu, metrics = 'all', start_date = None,
         elif i == 'NSE':
             val = ofs.nashsutcliffe(df['obs'].values, df['simu'].values)
         elif i == 'logNSE':
+            # default epsilon is 1/100 of the mean observed flow
             if epsilon == 0:
                 epsilon = df['obs'].mean() / 100
             val = ofs.lognashsutcliffe(df['obs'].values, df['simu'].values, epsilon=epsilon)
@@ -501,3 +581,23 @@ def array2asc(file_array, file_asc, nx=600, ny=600, dx=100, model_origin = [0,0]
     print('Exported ASCII file is in {}'.format(file_asc))
     
     return None
+
+def mann_whitney_test(data1, data2, alpha=0.05):
+    """
+    perform a Mann-Whitney test, and return statistics and p-value
+    Parameters:
+        x: data1, dataframe timeseries
+        y: data2, dataframe timeseries
+        alpha: significance factor, default 0.05
+    Returns:
+        stat, p
+    """
+    stat, p = mannwhitneyu(data1, data2)
+    print('Statistics=%.3f, p=%.3f' % (stat, p))
+    if p > alpha:
+        print('Same distributions (fail to reject H0)')
+    else:
+        print('Different distributions (reject H0)')
+    return stat, p
+        
+    return stat, p
